import torch
import torch.nn  as nn
from math import sqrt


def norm_1d(x):
    eps = 1e-5
    gamma = torch.ones(x.shape[1])
    beta = torch.zeros(x.shape[1])
    x_mean = torch.mean(x, dim=0, keepdim=True)
    x_var = torch.mean((x - x_mean) ** 2, dim=0, keepdim=True)
    x_hat = (x - x_mean) / torch.sqrt(x_var + eps)
    return gamma.view_as(x_mean) * x_hat + beta.view_as(x_mean)


class EqualLR:
    def __init__(self, name):
        self.name = name

    def compute_weight(self, module):
        weight = getattr(module, self.name + '_orig')
        fan_in = weight.data.size(1) * weight.data[0][0].numel()

        return weight * sqrt(2 / fan_in)

    @staticmethod
    def apply(module, name):
        fn = EqualLR(name)

        weight = getattr(module, name)
        del module._parameters[name]
        module.register_parameter(name + '_orig', nn.Parameter(weight.data))
        module.register_forward_pre_hook(fn)

        return fn

    def __call__(self, module, input):
        weight = self.compute_weight(module)
        setattr(module, self.name, weight)


def equal_lr(module, name='weight'):
    EqualLR.apply(module, name)

    return module


class EqualLinear(nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()

        linear = nn.Linear(in_dim, out_dim)
        linear.weight.data.normal_()
        linear.bias.data.zero_()

        self.linear = equal_lr(linear)

    def forward(self, input):
        return self.linear(input)


class AdaptiveInstanceNorm(nn.Module):
    def __init__(self, in_channel, style_dim):
        super().__init__()

        self.norm = nn.InstanceNorm2d(in_channel)
        self.style = EqualLinear(style_dim, in_channel * 2)

        self.style.linear.bias.data[:in_channel] = 1
        self.style.linear.bias.data[in_channel:] = 0

    def forward(self, input, style):
        style = self.style(style).unsqueeze(2).unsqueeze(3)
        gamma, beta = style.chunk(2, 1)

        out = self.norm(input)
        out = gamma * out + beta

        return out


class ResnetBlock(nn.Module):
    def __init__(self, dim, norm_layer, use_bias=False):
        super(ResnetBlock, self).__init__()
        self.conv_block = self.build_conv_block(dim, norm_layer, use_bias)

    def build_conv_block(self, dim, norm_layer, use_bias):
        conv_block = []
        conv_block += [nn.ReflectionPad2d(1)]
        conv_block += [nn.Conv2d(dim, dim, kernel_size=3, padding=0, bias=use_bias), norm_layer(dim), nn.ReLU(True)]
        conv_block += [nn.ReflectionPad2d(1)]
        conv_block += [nn.Conv2d(dim, dim, kernel_size=3, padding=0, bias=use_bias), norm_layer(dim)]
        return nn.Sequential(*conv_block)

    def forward(self, x):
        out = x + self.conv_block(x)
        return out
    
    
# Optical Flow Encoder
class OFE(nn.Module):
    def __init__(self, ngf, dlatent):
        super(OFE, self).__init__()

        self.downConv1 = nn.Conv3d(2, ngf, (3, 4, 4), stride=(1, 2, 2), padding=(1, 1, 1), bias=False)
        self.downConv2 = nn.Conv3d(ngf, ngf * 2, (4, 4, 4), stride=(2, 2, 2), padding=(1, 1, 1), bias=False)
        self.downConv3 = nn.Conv3d(ngf * 2, ngf * 4, 4, 2, 1, bias=False)
        self.downConv4 = nn.Conv3d(ngf * 4, ngf * 8, 4, 2, 1, bias=False)
        self.downConv5 = nn.Conv3d(ngf * 8, ngf * 16, 4, 2, 1, bias=False)
        self.downConv6 = nn.Conv3d(ngf * 16, dlatent, (2, 4, 4), stride=(1, 1, 1), padding=(0, 0, 0), bias=False)

        self.downBN2 = nn.BatchNorm3d(ngf * 2)
        self.downBN3 = nn.BatchNorm3d(ngf * 4)
        self.downBN4 = nn.BatchNorm3d(ngf * 8)
        self.downBN5 = nn.BatchNorm3d(ngf * 16)
        self.lrelu = nn.LeakyReLU(0.2, inplace=True)

    def forward(self, x):
        downx1 = self.downConv1(x)
        downx2 = self.downConv2(downx1)
        downx2 = self.downBN2(downx2)
        downx2 = self.lrelu(downx2)
        downx3 = self.downConv3(downx2)
        downx3 = self.downBN3(downx3)
        downx3 = self.lrelu(downx3)
        downx4 = self.downConv4(downx3)
        downx4 = self.downBN4(downx4)
        downx4 = self.lrelu(downx4)
        downx5 = self.downConv5(downx4)
        downx5 = self.downBN5(downx5)
        downx5 = self.lrelu(downx5)
        downx6 = self.downConv6(downx5)
        b, s = downx6.size(0), downx6.size(1)
        out = downx6.view(b, s)
        out = norm_1d(out)

        return out


# Content Stream
class ContentStream(nn.Module):
    def __init__(self, ngf, n_blocks):
        super(ContentStream, self).__init__()
        self.ngf = ngf
        self.n_blocks = n_blocks

        content_stream = []
        for i in range(self.n_blocks):
            content_stream += [ResnetBlock(ngf * 2, norm_layer=nn.InstanceNorm2d, use_bias=False)]
        self.content_stream = nn.Sequential(*content_stream)

    def forward(self, x):
        out = self.content_stream(x)
        return out


# Motion Stream
class MotionStream(nn.Module):
    def __init__(self, ngf, dlatent, n_blocks=6):
        super(MotionStream, self).__init__()
        self.ngf = ngf
        self.dlatent = dlatent
        self.n_blocks = n_blocks

        # solve input image
        conv = [nn.Conv2d(ngf * 2, ngf, kernel_size=3, stride=1, padding=1, bias=False),
                nn.InstanceNorm2d(ngf),
                nn.ReLU(True)]
        self.conv = nn.Sequential(*conv)
        # solve input vector
        n_mlp = 3
        motion = []
        for i in range(n_mlp):
            motion.append(EqualLinear(dlatent, dlatent))
            motion.append(nn.LeakyReLU(0.2))

        model_img = []
        model_A = []
        model_AdaIN = []
        for i in range(n_blocks):
            model_img += [ResnetBlock(ngf, norm_layer=nn.InstanceNorm2d, use_bias=False)]
            model_A += [nn.Sequential(EqualLinear(dlatent, dlatent), nn.LeakyReLU(0.2),
                                      EqualLinear(dlatent, dlatent), nn.LeakyReLU(0.2))]
            model_AdaIN += [AdaptiveInstanceNorm(ngf, dlatent)]

        model_out = [nn.Conv2d(ngf, 2 * 32, kernel_size=3, stride=1, dilation=1, padding=1, bias=True),
                     nn.LeakyReLU(0.1, inplace=True)]

        self.motion = nn.Sequential(*motion)
        self.model_img = nn.Sequential(*model_img)
        self.model_A = nn.Sequential(*model_A)
        self.model_AdaIN = nn.Sequential(*model_AdaIN)
        self.model_out = nn.Sequential(*model_out)

    def forward(self, x, f):
        x = self.conv(x)
        f = self.motion(f)
        for i in range(self.n_blocks):
            x = self.model_img[i](x)
            f_i = self.model_A[i](f)
            x = self.model_AdaIN[i](x, f_i)
        out = self.model_out(x)
        return out


# DTVNet
class DTVNet(nn.Module):
    def __init__(self, ngf=64, dlatent=512, n_blocks=9):
        super(DTVNet, self).__init__()

        self.ngf = ngf
        self.dlatent = dlatent
        self.n_blocks = n_blocks

        # OFE
        self.flow_encoder = OFE(self.ngf, self.dlatent)
        # shared encoder
        shared_encoder = [nn.ReflectionPad2d(2),
                          nn.Conv2d(3, ngf, kernel_size=5, stride=1, padding=0, bias=False),
                          nn.InstanceNorm2d(ngf),
                          nn.ReLU(True)]
        shared_encoder += [nn.Conv2d(ngf * 1, ngf * 2, kernel_size=3, stride=2, padding=1, bias=False),
                           nn.InstanceNorm2d(ngf * 2),
                           nn.ReLU(True)]
        shared_encoder += [nn.Conv2d(ngf * 2, ngf * 2, kernel_size=3, stride=1, padding=1, bias=False),
                           nn.InstanceNorm2d(ngf * 2),
                           nn.ReLU(True)]
        self.shared_encoder = nn.Sequential(*shared_encoder)
        # content stream
        self.content_stream = ContentStream(ngf, n_blocks)
        # motion stream
        self.motion_stream = MotionStream(ngf, dlatent)
        # decoder
        ngf_out = 96
        decoder = []
        decoder += [nn.Conv2d(ngf * 3, ngf_out, kernel_size=3, stride=1, padding=1, bias=False),
                    nn.InstanceNorm2d(ngf_out),
                    nn.ReLU(True)]
        decoder += [nn.Conv2d(ngf_out, ngf_out, kernel_size=3, stride=1, padding=1, bias=False),
                    nn.InstanceNorm2d(ngf_out),
                    nn.ReLU(True)]
        decoder += [nn.ConvTranspose2d(ngf_out, ngf_out, kernel_size=3, stride=2, padding=1, output_padding=1, bias=False),
                    nn.InstanceNorm2d(ngf_out),
                    nn.ReLU(True)]
        decoder += [nn.ReflectionPad2d(2),
                    nn.Conv2d(ngf_out, ngf_out, kernel_size=5, padding=0),
                    nn.Tanh()]
        self.decoder = nn.Sequential(*decoder)

    def forward(self, x, flow):
        f = flow
        if self.training:
            f = self.flow_encoder(f)
        # convs
        x_share = self.shared_encoder(x)                           # [B, 128, 64, 64]
        x_content = self.content_stream(x_share)                    # [B, 128, 64, 64]
        x_motion = self.motion_stream(x_share, f)           # [B, 32*2, 64, 64]
        x_cat = torch.cat([x_content, x_motion], dim=1)    # [B, 32*6, 64, 64]
        out = self.decoder(x_cat)  # [B, 32*2, 64, 64]
        out = out.view(out.shape[0], 3, 32, out.shape[2], out.shape[3])

        return out, x_motion


# Discriminator
class Discriminator(nn.Module):
    def __init__(self, ndf):
        super(Discriminator, self).__init__()
        self.slice1 = nn.Sequential(
            nn.Conv3d(3, ndf, (3, 4, 4), stride=(1, 2, 2), padding=(1, 1, 1), bias=False),
            nn.LeakyReLU(0.2, inplace=True),
        )
        self.slice2 = nn.Sequential(
            nn.Conv3d(ndf, ndf * 2, (4, 4, 4), stride=(2, 2, 2), padding=(1, 1, 1), bias=False),
            nn.InstanceNorm3d(ndf * 2),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv3d(ndf * 2, ndf * 4, 4, 2, 1, bias=False),
            nn.InstanceNorm3d(ndf * 4),
            nn.LeakyReLU(0.2, inplace=True),
        )
        self.slice3 = nn.Sequential(
            nn.Conv3d(ndf * 4, ndf * 8, 4, 2, 1, bias=False),
            nn.InstanceNorm3d(ndf * 8),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv3d(ndf * 8, ndf * 16, 4, 2, 1, bias=False),
            nn.InstanceNorm3d(ndf * 16),
            nn.LeakyReLU(inplace=True),
        )
        self.slice4 = nn.Sequential(
            nn.Conv3d(ndf * 16, 1, (2, 4, 4), stride=(1, 1, 1), padding=(0, 0, 0), bias=False),
        )

    def forward(self, x):  # [B, 3, 32, 128, 128]
        x1 = self.slice1(x)  # [B, 128, 32, 32, 32]
        x2 = self.slice2(x1)  # [B, 512, 8, 8, 8]
        x3 = self.slice3(x2)  # [B, 2048, 2, 2, 2]
        x4 = self.slice4(x3)  # [B, 1, 1, 1, 1]
        return x4.view(-1, 1)


if __name__ == '__main__':
    input = torch.randn(2, 3, 128, 128)
    flow = torch.randn(2, 2, 32, 128, 128)
    flow_vector = torch.randn(2, 512)
    net = DTVNet(ngf=64, dlatent=512, n_blocks=9)
    # train
    net.train()
    imgs, flows = net(input, flow)
    print('imgs:{}'.format(imgs.shape))
    print('flows:{}'.format(flows.shape))
    # test
    net.eval()
    imgs, flows = net(input, flow_vector)
    print('imgs:{}'.format(imgs.shape))
    print('flows:{}'.format(flows.shape))

